<?php

namespace FiloBlu\OrderStatus\Model\Rewrite;

use Exception;
use Magento\Paypal\Model\Info;

/**
 * FiloBlu customized the following methods:
 * - _registerPaymentCapture()
 * - _registerPaymentAuthorization()
 */
class Ipn extends \Magento\Paypal\Model\Ipn implements \Magento\Paypal\Model\IpnInterface
{
    
    const IPN_COMPLETED_ORDER_STATUS = 'payment_confirmed';
    const IPN_AUTHORIZED_ORDER_STATUS = 'confirmed';
    const AUTHORIZED_ORDER_STATUS_CONFIG_PATH = 'payment/adyen_abstract/payment_pre_authorized';
    const CAPTURED_ORDER_STATUS_CONFIG_PATH = 'payment/adyen_abstract/payment_authorized';

    /**
     * Get ipn data, send verification to PayPal, run corresponding handler
     *
     * @return void
     * @throws Exception
     */
    public function processIpnRequest()
    {
        
        $this->_addDebugData('ipn', $this->getRequestData());

        try {
            $this->_getConfig();
            $this->_postBack();
            $this->_processOrder();
        } catch (Exception $e) {
            $this->_addDebugData('exception', $e->getMessage());
            $this->_debug();
            throw $e;
        }
        $this->_debug();
    }
    
    
    /**
     * IPN workflow implementation
     * Everything should be added to order comments. In positive processing cases customer will get email notifications.
     * Admin will be notified on errors.
     *
     * @return void
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    protected function _processOrder()
    {
        $this->_getConfig();
        try {
            // Handle payment_status
            $transactionType = $this->getRequestData('txn_type');
            switch ($transactionType) {
                // handle new case created
                case Info::TXN_TYPE_NEW_CASE:
                    $this->_registerDispute();
                    break;
                    // handle new adjustment is created
                case Info::TXN_TYPE_ADJUSTMENT:
                    $this->_registerAdjustment();
                    break;
                    //handle new transaction created
                default:
                    $this->_registerTransaction();
                    break;
            }
        } catch (\Magento\Framework\Exception\LocalizedException $e) {
            $comment = $this->_createIpnComment(__('Note: %1', $e->getMessage()), true);
            $comment->save();
            throw $e;
        }
    }
    
    
    /**
     * Process regular IPN notifications
     *
     * @return void
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws Exception
     * @SuppressWarnings(PHPMD.CyclomaticComplexity)
     */
    protected function _registerTransaction()
    {
        // Handle payment_status
        $paymentStatus = $this->_filterPaymentStatus($this->getRequestData('payment_status'));
        switch ($paymentStatus) {
            // paid
            case Info::PAYMENTSTATUS_COMPLETED:
                $this->_registerPaymentCapture(true);
                break;
                // the holded payment was denied on paypal side
            case Info::PAYMENTSTATUS_DENIED:
                $this->_registerPaymentDenial();
                break;
                // customer attempted to pay via bank account, but failed
            case Info::PAYMENTSTATUS_FAILED:
                // cancel order
                $this->_registerPaymentFailure();
                break;
                // payment was obtained, but money were not captured yet
            case Info::PAYMENTSTATUS_PENDING:
                $this->_registerPaymentPending();
                break;
            case Info::PAYMENTSTATUS_PROCESSED:
                $this->_registerMasspaymentsSuccess();
                break;
            case Info::PAYMENTSTATUS_REVERSED:
                //break is intentionally omitted
            case Info::PAYMENTSTATUS_UNREVERSED:
                $this->_registerPaymentReversal();
                break;
            case Info::PAYMENTSTATUS_REFUNDED:
                $this->_registerPaymentRefund();
                break;
                // authorization expire/void
            case Info::PAYMENTSTATUS_EXPIRED:
                // break is intentionally omitted
            case Info::PAYMENTSTATUS_VOIDED:
                $this->_registerPaymentVoid();
                break;
            default:
                throw new Exception("Cannot handle payment status '{$paymentStatus}'.");
        }
    }
    
    
    /**
     * Process completed payment (either full or partial)
     *
     * @param bool $skipFraudDetection
     * @return void
     */
    protected function _registerPaymentCapture($skipFraudDetection = false)
    {
        if ($this->getRequestData('transaction_entity') == 'auth') {
            return;
        }
        $parentTransactionId = $this->getRequestData('parent_txn_id');
        $this->_importPaymentInformation();
        $payment = $this->_order->getPayment();
        $payment->setTransactionId(
            $this->getRequestData('txn_id')
        );
        $payment->setCurrencyCode(
            $this->getRequestData('mc_currency')
        );
        $payment->setPreparedMessage(
            $this->_createIpnComment('')
        );
        $payment->setParentTransactionId(
            $parentTransactionId
        );
        $payment->setShouldCloseParentTransaction(
            'Completed' === $this->getRequestData('auth_status')
        );
        $payment->setIsTransactionClosed(
            0
        );
        $payment->registerCaptureNotification(
            $this->getRequestData('mc_gross'),
            $skipFraudDetection && $parentTransactionId
        );
        
        //FiloBlu customization begin
        $captured_order_status = $this->_config->getValue(self::CAPTURED_ORDER_STATUS_CONFIG_PATH, \Magento\Store\Model\ScopeInterface::SCOPE_STORE) ? 
                        $this->_config->getValue(self::CAPTURED_ORDER_STATUS_CONFIG_PATH, \Magento\Store\Model\ScopeInterface::SCOPE_STORE) : self::IPN_COMPLETED_ORDER_STATUS;
        $this->_order->addStatusHistoryComment(
            __('Order processed by mod2-order-status'), $captured_order_status
        );
        //FiloBlu customization end
        
        $this->_order->save();

        // notify customer
        $invoice = $payment->getCreatedInvoice();
        if ($invoice && !$this->_order->getEmailSent()) {
            $this->orderSender->send($this->_order);
            $this->_order->addStatusHistoryComment(
                __('You notified customer about invoice #%1.', $invoice->getIncrementId())
            )->setIsCustomerNotified(
                true
            )->save();
        }
    }
    
    
    /**
     * Process payment pending notification
     *
     * @return void
     * @throws Exception
     */
    public function _registerPaymentPending()
    {
        $reason = $this->getRequestData('pending_reason');
        if ('authorization' === $reason) {
            $this->_registerPaymentAuthorization();
            return;
        }
        if ('order' === $reason) {
            throw new Exception('The "order" authorizations are not implemented.');
        }
        // case when was placed using PayPal standard
        if (\Magento\Sales\Model\Order::STATE_PENDING_PAYMENT == $this->_order->getState()
            && !$this->getRequestData('transaction_entity')
        ) {
            $this->_registerPaymentCapture();
            return;
        }

        $this->_importPaymentInformation();

        $this->_order->getPayment()->setPreparedMessage(
            $this->_createIpnComment($this->_paypalInfo->explainPendingReason($reason))
        )->setTransactionId(
            $this->getRequestData('txn_id')
        )->setIsTransactionClosed(
            0
        )->update(false);
        $this->_order->save();
    }

    /**
     * Register authorized payment
     *
     * @return void
     */
    protected function _registerPaymentAuthorization()
    {
        /** @var $payment \Magento\Sales\Model\Order\Payment */
        $payment = $this->_order->getPayment();
        if ($this->_order->canFetchPaymentReviewUpdate()) {
            $payment->update(true);
        } else {
            $this->_importPaymentInformation();
            $payment->setPreparedMessage(
                $this->_createIpnComment('')
            )->setTransactionId(
                $this->getRequestData('txn_id')
            )->setParentTransactionId(
                $this->getRequestData('parent_txn_id')
            )->setCurrencyCode(
                $this->getRequestData('mc_currency')
            )->setIsTransactionClosed(
                0
            )->registerAuthorizationNotification(
                $this->getRequestData('mc_gross')
            );
        }
        if (!$this->_order->getEmailSent()) {
            $this->orderSender->send($this->_order);
        }
        
        //FiloBlu customization begin
        $authorized_order_status = $this->_config->getValue(self::AUTHORIZED_ORDER_STATUS_CONFIG_PATH, \Magento\Store\Model\ScopeInterface::SCOPE_STORE) ? 
                        $this->_config->getValue(self::AUTHORIZED_ORDER_STATUS_CONFIG_PATH, \Magento\Store\Model\ScopeInterface::SCOPE_STORE) : self::IPN_AUTHORIZED_ORDER_STATUS;
        $this->_order->addStatusHistoryComment(
            __('Order processed by mod2-order-status'), $authorized_order_status
        );
        //FiloBlu customization end
        
        $this->_order->save();
    }
    
}
